// Copyright (C) 2005 Christian Mendl
//
// standard dynamic array, using 'placement new' to construct objects

#ifndef ARRAY_H
#define ARRAY_H

#include <new>
#include <assert.h>
#include "next2k.h"


template<class T>
class Array
{

public:

	Array();
	Array(int _size);
	Array(const T *_data, int _size);
	Array(const Array& that);

	~Array();

	Array& operator= (const Array& that);

	T& operator[] (int index);
	const T& operator[] (int index) const;

	// return data pointer
	T* Data();
	const T *Data() const;

	int Size() const;
	bool Empty() const;
	int NumBytes() const;

	T& Front();
	const T& Front() const;

	T& Back();
	const T& Back() const;

	// returns index of first element equal to 't', or -1 if 't' can't be found
	int Find(const T& t) const;

	void Resize(int _size);

	void Grow(int numGrow);

	void Shrink(int numShrink);

	void PushBack(const T& t);

	template<class S, template<class> class Ar>
	void PushBack(const Ar<S>& that);

	template<class S>
	void PushBack(const S *v, int num);

	void PopBack();

	// remove element at index 'index'
	void EraseAt(int index);

	// look for 't' and remove first occurrence; return false if 't' can't be found
	bool Erase(const T& t);

	void Clear();

private:

	T *data;
	int size, capacity;
};


template<class T>
inline Array<T>::Array()
{
	size = 0;

	capacity = 4;
	data = (T *)malloc(capacity*sizeof(T));
}

template<class T>
inline Array<T>::Array(int _size)
{
	capacity = size = _size;

	data = (T *)malloc(capacity*sizeof(T));

	for (int i = 0; i < size; i++)
		new(data+i) T();
}

template<class T>
inline Array<T>::Array(const T *_data, int _size)
{
	capacity = size = _size;

	data = (T *)malloc(capacity*sizeof(T));

	for (int i = 0; i < size; i++)
		new(data+i) T(_data[i]);
}

template<class T>
inline Array<T>::Array(const Array& that)
{
	capacity = size = that.size;

	data = (T *)malloc(capacity*sizeof(T));

	for (int i = 0; i < size; i++)
		new(data+i) T(that.data[i]);
}


template<class T>
inline Array<T>::~Array()
{
	for (int i = size-1; i >= 0; i--)
		(data+i)->~T();

	free(data);
}


template<class T>
inline Array<T>& Array<T>::operator= (const Array& that)
{
	for (int i = size-1; i >= 0; i--)
		(data+i)->~T();

	size = that.size;

	if (capacity < size)
	{
		capacity = size;

		// this should be faster than realloc
		free(data);
		data = (T *)malloc(capacity*sizeof(T));
	}

	for (int i = 0; i < size; i++)
		new(data+i) T(that.data[i]);

	return *this;
}


template<class T>
inline T& Array<T>::operator[] (int index)
{
	assert(0 <= index && index < size);

	return data[index];
}

template<class T>
inline const T& Array<T>::operator[] (int index) const
{
	assert(0 <= index && index < size);

	return data[index];
}


template<class T>
inline T *Array<T>::Data()
{
	return data;
}

template<class T>
inline const T *Array<T>::Data() const
{
	return data;
}


template<class T>
inline int Array<T>::Size() const
{
	return size;
}

template<class T>
inline bool Array<T>::Empty() const
{
	return size == 0;
}

template<class T>
inline int Array<T>::NumBytes() const
{
	return size * sizeof(T);
}


template<class T>
inline T& Array<T>::Front()
{
	assert(size > 0);

	return data[0];
}

template<class T>
inline const T& Array<T>::Front() const
{
	assert(size > 0);

	return data[0];
}


template<class T>
inline T& Array<T>::Back()
{
	assert(size > 0);

	return data[size-1];
}

template<class T>
inline const T& Array<T>::Back() const
{
	assert(size > 0);

	return data[size-1];
}


template<class T>
inline int Array<T>::Find(const T& t) const
{
	for (int i = 0; i < size; i++)
		if (data[i] == t)
			return i;

	return -1;
}


template<class T>
inline void Array<T>::Resize(int _size)
{
	assert(_size >= 0);

	if (size < _size)
	{
		if (capacity < _size)
		{
			capacity = _size;
			data = (T *)realloc(data, capacity*sizeof(T));
		}

		for (int i = size; i < _size; i++)
			new(data+i) T();
	}
	else
	{
		for (int i = size-1; i >= _size; i--)
			(data+i)->~T();
	}

	size = _size;
}


template<class T>
inline void Array<T>::Grow(int numGrow)
{
	assert(numGrow >= 0);

	size += numGrow;

	if (capacity < size)
	{
		capacity = Next2K(size);
		data = (T *)realloc(data, capacity*sizeof(T));
	}

	for (int i = size-numGrow; i < size; i++)
		new(data+i) T();
}


template<class T>
inline void Array<T>::Shrink(int numShrink)
{
	assert(numShrink >= 0 && size-numShrink >= 0);

	for (int i = size-1; i >= size-numShrink; i--)
		(data+i)->~T();

	size -= numShrink;
}


template<class T>
inline void Array<T>::PushBack(const T& t)
{
	++size;

	if (capacity < size)
	{
		capacity = Next2K(size);
		data = (T *)realloc(data, capacity*sizeof(T));
	}

	new(data+size-1) T(t);
}

template<class T>
template<class S, template<class> class Ar>
inline void Array<T>::PushBack(const Ar<S>& that)
{
	if (capacity < size+that.Size())
	{
		capacity = Next2K(size+that.Size());
		data = (T *)realloc(data, capacity*sizeof(T));
	}

	for (int i = 0; i < that.Size(); i++)
		new(data+size+i) T(that[i]);

	size += that.Size();
}

template<class T>
template<class S>
inline void Array<T>::PushBack(const S *v, int num)
{
	if (capacity < size+num)
	{
		capacity = Next2K(size+num);
		data = (T *)realloc(data, capacity*sizeof(T));
	}

	for (int i = 0; i < num; i++)
		new(data+size+i) T(v[i]);

	size += num;
}


template<class T>
inline void Array<T>::PopBack()
{
	assert(size > 0);

	--size;
	(data+size)->~T();
}


template<class T>
inline void Array<T>::EraseAt(int index)
{
	assert(0 <= index && index < size);

	// shift elements by one
	--size;
	for (int i = index; i < size; i++)
		data[i] = data[i+1];

	// delete last element
	(data+size)->~T();
}

template<class T>
inline bool Array<T>::Erase(const T& t)
{
	int index = Find(t);

	if (index >= 0)
	{
		EraseAt(index);
		return true;
	}

	return false;
}


template<class T>
inline void Array<T>::Clear()
{
	for (int i = size-1; i >= 0; i--)
		(data+i)->~T();

	size = 0;
}



#endif
