// Copyright (C) 2005 Christian Mendl
//
// open addressing hash table with linear probing and single-slot stepping;
// objects of class H must provide a hash function by defining 'operator()'

#ifndef HASH_SET_H
#define HASH_SET_H

#include <assert.h>


template<class K, class H>
class HashSet
{

public:

	struct Node;

	class Iterator
	{

	public:

		Iterator();
		Iterator(const Node *_n);

		const K& operator* () const;
		const K *operator-> () const;

		Iterator& operator++ ();
		Iterator& operator-- ();

		bool operator== (const Iterator& that) const;
		bool operator!= (const Iterator& that) const;

	private:

		friend class HashSet;

		const Node *n;
	};

	HashSet();
	HashSet(int _capacity);
	HashSet(const HashSet& that);

	~HashSet();

	HashSet& operator= (const HashSet& that);

	Iterator Begin() const;
	Iterator ReverseBegin() const;
	Iterator End() const;
	Iterator ReverseEnd() const;

	// return 'End()' if key is not in the table
	Iterator Find(const K& k) const;

	bool Contains(const K& k) const;

	int Size() const;
	bool Empty() const;

	// return 'End()' on successfull insertion,
	// otherwise pointer to node containing 'k'
	Iterator Insert(const K& k);

	// return true if 'k' has been erased;
	// if 'k' isn't in the hash set, it will remain unchanged
	bool Erase(const K& k);

	// point 'it' to the next node
	void Erase(Iterator& it);

	void Clear();

private:

	// rebuild larger table if
	// size > REBUILD_LIMIT/100 * capacity
	static const int REBUILD_LIMIT = 75;

	struct Node
	{
		K k;		// key
		bool free;	// slot occupied or not?

		Node(const K& _k)
		: k(_k), free(false) {
		}

		Node(const Node& that)
		: k(that.k), free(that.free) {
		}
	};

	// add 2 additional "occupied" nodes at the beginning and the end
	// of the array to make sure iterators stay in range,
	// 'capacity' doesn't count these nodes;
	//
	// at least 1 node must be free, i.e. size <= capacity - 1
	//
	// for every key stored, there must be no free nodes between the hash position and actual position

	Node *nodes;
	int size, capacity;

	const H hash;	// the hash function object

	void Rebuild();
	void Erase(int i);	// 'i' is node index
};


template<class K, class H>
inline HashSet<K, H>::HashSet()
{
	size = 0;

	capacity = 4;
	nodes = ((Node *)malloc((capacity+2)*sizeof(Node)))+1;	// point to first used node
	nodes[-1].free = nodes[capacity].free = false;
	for (int i = 0; i < capacity; i++)
		nodes[i].free = true;
}

template<class K, class H>
inline HashSet<K, H>::HashSet(int _capacity)
{
	size = 0;

	capacity = _capacity;
	nodes = ((Node *)malloc((capacity+2)*sizeof(Node)))+1;
	nodes[-1].free = nodes[capacity].free = false;
	for (int i = 0; i < capacity; i++)
		nodes[i].free = true;
}

template<class K, class H>
inline HashSet<K, H>::HashSet(const HashSet& that)
{
	size = that.size;

	capacity = that.capacity;
	nodes = ((Node *)malloc((capacity+2)*sizeof(Node)))+1;
	nodes[-1].free = nodes[capacity].free = false;
	for (int i = 0; i < capacity; i++)
	{
		if (that.nodes[i].free)
			nodes[i].free = true;
		else
			new(nodes+i) Node(that.nodes[i]);
	}
}


template<class K, class H>
inline HashSet<K, H>::~HashSet()
{
	for (int i = 0; i < capacity; i++)
	{
		if (!nodes[i].free)
			nodes[i].k.~K();
	}
	free(nodes-1);
}


template<class K, class H>
inline HashSet<K, H>& HashSet<K, H>::operator= (const HashSet& that)
{
	for (int i = 0; i < capacity; i++)
	{
		if (!nodes[i].free)
			nodes[i].k.~K();
	}
	free(nodes-1);

	size = that.size;

	capacity = that.capacity;
	nodes = ((Node *)malloc((capacity+2)*sizeof(Node)))+1;
	nodes[-1].free = nodes[capacity].free = false;
	for (int i = 0; i < capacity; i++)
	{
		if (that.nodes[i].free)
			nodes[i].free = true;
		else
			new(nodes+i) Node(that.nodes[i]);
	}

	return *this;
}


template<class K, class H>
inline typename HashSet<K, H>::Iterator HashSet<K, H>::Begin() const
{
	int i = 0;
	while (nodes[i].free)	// using sentinel node at the end
		i++;

	return Iterator(nodes+i);
}

template<class K, class H>
inline typename HashSet<K, H>::Iterator HashSet<K, H>::ReverseBegin() const
{
	int i = capacity-1;
	while (nodes[i].free)
		i--;

	return Iterator(nodes+i);
}

template<class K, class H>
inline typename HashSet<K, H>::Iterator HashSet<K, H>::End() const
{
	return Iterator(nodes+capacity);
}

template<class K, class H>
inline typename HashSet<K, H>::Iterator HashSet<K, H>::ReverseEnd() const
{
	return Iterator(nodes-1);
}


template<class K, class H>
inline typename HashSet<K, H>::Iterator HashSet<K, H>::Find(const K& k) const
{
	int i = hash(k) % capacity;
	assert(i >= 0);

	while (!nodes[i].free)	// at least 1 node is free
	{
		if (nodes[i].k == k)
			return Iterator(nodes+i);

		i = (i+1) % capacity;
	}

	return Iterator(nodes+capacity);
}


template<class K, class H>
inline bool HashSet<K, H>::Contains(const K& k) const
{
	int i = hash(k) % capacity;
	assert(i >= 0);

	while (!nodes[i].free)	// at least 1 node is free
	{
		if (nodes[i].k == k)
			return true;

		i = (i+1) % capacity;
	}

	return false;
}


template<class K, class H>
inline int HashSet<K, H>::Size() const
{
	return size;
}

template<class K, class H>
inline bool HashSet<K, H>::Empty() const
{
	return size == 0;
}


template<class K, class H>
inline typename HashSet<K, H>::Iterator HashSet<K, H>::Insert(const K& k)
{
	int i = hash(k) % capacity;
	assert(i >= 0);

	while (!nodes[i].free)	// at least 1 node is free
	{
		if (nodes[i].k == k)
			return Iterator(nodes+i);

		i = (i+1) % capacity;
	}

	// key not in table yet

	++size;

	if (100*size > REBUILD_LIMIT*capacity)
	{
		Rebuild();

		i = hash(k) % capacity;

		while (!nodes[i].free)
		{
			assert(nodes[i].k != k);

			i = (i+1) % capacity;
		}
	}

	return Iterator(new(nodes+i) Node(k));
}


template<class K, class H>
inline bool HashSet<K, H>::Erase(const K& k)
{
	int i = hash(k) % capacity;
	assert(i >= 0);

	while (!nodes[i].free)
	{
		if (nodes[i].k == k)
		{
			// found key
			Erase(i);
			return true;
		}

		i = (i+1) % capacity;
	}

	// key not in table
	return false;
}

template<class K, class H>
inline void HashSet<K, H>::Erase(Iterator& it)
{
	int i = int(it.n - nodes);

	Erase(i);

	while (nodes[i].free)	// using sentinel node at the end
		i++;
	it.n = nodes + i;
}


template<class K, class H>
inline void HashSet<K, H>::Clear()
{
	size = 0;

	for (int i = 0; i < capacity; i++)
		if (!nodes[i].free)
		{
			nodes[i].k.~K();
			nodes[i].free = true;
		}
}


template<class K, class H>
inline void HashSet<K, H>::Rebuild()
{
	Node *nodesOld = nodes;

	// allocate new nodes
	capacity *= 2;
	nodes = ((Node *)malloc((capacity+2)*sizeof(Node)))+1;
	nodes[-1].free = nodes[capacity].free = false;
	for (int i = 0; i < capacity; i++)
		nodes[i].free = true;

	// reinsert keys
	for (int i = 0; i < capacity/2; i++)
	{
		if (!nodesOld[i].free)
		{
			int j = hash(nodesOld[i].k) % capacity;
			assert(j >= 0);

			while (!nodes[j].free)	// at least 1 node is free
			{
				assert(nodes[j].k != nodesOld[i].k);

				j = (j+1) % capacity;
			}

			new(nodes+j) Node(nodesOld[i]);
		}
	}

	// delete old nodes
	for (int i = 0; i < capacity/2; i++)
	{
		if (!nodesOld[i].free)
			nodesOld[i].k.~K();
	}
	free(nodesOld-1);
}


template<class K, class H>
inline void HashSet<K, H>::Erase(int i)
{
	assert(0 <= i && i < capacity);
	assert(!nodes[i].free);

	// no free nodes between the hash position and actual position allowed

	int j = (i+1) % capacity;
	while (!nodes[j].free)
	{
		assert(i != j);

		int m = hash(nodes[j].k) % capacity;
		if (j > i)
		{
			if (m <= i || j < m)
			{
				nodes[i].k = nodes[j].k;
				i = j;
			}
		}
		else
		{
			if (j < m && m <= i)
			{
				nodes[i].k = nodes[j].k;
				i = j;
			}
		}

		j = (j+1) % capacity;
	}

	nodes[i].k.~K();
	nodes[i].free = true;

	--size;
}


template<class K, class H>
inline HashSet<K, H>::Iterator::Iterator() {
}

template<class K, class H>
inline HashSet<K, H>::Iterator::Iterator(const Node *_n)
: n(_n) {
}


template<class K, class H>
inline const K& HashSet<K, H>::Iterator::operator* () const
{
	return n->k;
}

template<class K, class H>
inline const K *HashSet<K, H>::Iterator::operator-> () const
{
	return &n->k;
}


template<class K, class H>
inline typename HashSet<K, H>::Iterator& HashSet<K, H>::Iterator::operator++ ()
{
	while ((++n)->free)
		;

	return *this;
}

template<class K, class H>
inline typename HashSet<K, H>::Iterator& HashSet<K, H>::Iterator::operator-- ()
{
	while ((--n)->free)
		;

	return *this;
}


template<class K, class H>
inline bool HashSet<K, H>::Iterator::operator== (const Iterator& that) const
{
	return n == that.n;
}

template<class K, class H>
inline bool HashSet<K, H>::Iterator::operator!= (const Iterator& that) const
{
	return n != that.n;
}



#endif
