// Copyright (C) 2005 Christian Mendl
//
// double-linked list

#ifndef LIST_H
#define LIST_H

#include <assert.h>


template<class T>
class List
{

public:

	struct Node;

	class ConstIterator
	{

	public:

		ConstIterator();
		ConstIterator(Node *_n);

		const T& operator* () const;
		const T *operator-> () const;

		ConstIterator& operator++ ();
		ConstIterator& operator-- ();

		bool operator== (const ConstIterator& that) const;
		bool operator!= (const ConstIterator& that) const;

	protected:

		friend class List;

		Node *n;
	};

	class Iterator : public ConstIterator
	{

	public:

		Iterator();
		Iterator(Node *_n);

		T& operator* () const;	// adapting STL list: return value isn't constant
		T *operator-> () const;

		Iterator& operator++ ();
		Iterator& operator-- ();
	};

	List();
	List(const List& that);

	~List();

	List& operator= (const List& that);

	int Size() const;
	bool Empty() const;

	T& Front();

	T& Back();

	Iterator Begin();
	ConstIterator Begin() const;

	Iterator End();
	ConstIterator End() const;

	Iterator ReverseBegin();
	ConstIterator ReverseBegin() const;

	Iterator ReverseEnd();
	ConstIterator ReverseEnd() const;

	Iterator Find(const T& t);
	ConstIterator Find(const T& t) const;

	void PushFront(const T& t);
	void PushBack(const T& t);

	void PopFront();
	void PopBack();

	// insert before 'it'
	void Insert(Iterator it, const T& t);

	// 'it' will point to the next node
	void Erase(Iterator& it);

	// look for 't' and remove first occurrence; return false if 't' can't be found
	bool Erase(const T& t);

	void Clear();

private:

	struct Node
	{
		T t;
		Node *prev, *next;

		Node() {
		}

		Node(const T& _t)
		: t(_t) {
		}

		Node(const T& _t, Node *_prev, Node *_next)
		: t(_t), prev(_prev), next(_next) {
		}
	};

	// use sentinel nodes to avoid 'if's when accessing list
	// "sentinel node begin": 'next' points to first node
	// "sentinel node end": 'prev' points to last node
	Node *snb, *sne;

	int size;
};


template<class T>
inline List<T>::List()
: size(0), snb(new Node), sne(new Node)
{
	snb->next = sne;
	sne->prev = snb;
}

template<class T>
inline List<T>::List(const List& that)
: size(that.size), snb(new Node), sne(new Node)
{
	Node *n = snb;

	for (Node *c = that.snb->next; c != that.sne; c = c->next)
	{
		n->next = new Node(c->t);
		n->next->prev = n;
		n = n->next;
	}

	n->next = sne;
	sne->prev = n;
}


template<class T>
inline List<T>::~List()
{
	Node *c = snb;
	do
	{
		c = c->next;
		delete c->prev;
	}
	while (c != sne);

	delete sne;
}


template<class T>
inline List<T>& List<T>::operator= (const List& that)
{
	// clear list

	Node *c = snb->next;
	while (c != sne)
	{
		c = c->next;
		delete c->prev;
	}

	// copy size and elements in 'that'

	size = that.size;

	Node *n = snb;

	for (c = that.snb->next; c != that.sne; c = c->next)
	{
		n->next = new Node(c->t);
		n->next->prev = n;
		n = n->next;
	}

	n->next = sne;
	sne->prev = n;


	return *this;
}


template<class T>
inline int List<T>::Size() const
{
	return size;
}

template<class T>
inline bool List<T>::Empty() const
{
	return size == 0;
}


template<class T>
inline T& List<T>::Front()
{
	assert(size > 0);

	return snb->next->t;
}


template<class T>
inline T& List<T>::Back()
{
	assert(size > 0);

	return sne->prev->t;
}


template<class T>
inline typename List<T>::Iterator List<T>::Begin()
{
	return Iterator(snb->next);
}

template<class T>
inline typename List<T>::ConstIterator List<T>::Begin() const
{
	return ConstIterator(snb->next);
}


template<class T>
inline typename List<T>::Iterator List<T>::End()
{
	return Iterator(sne);
}

template<class T>
inline typename List<T>::ConstIterator List<T>::End() const
{
	return ConstIterator(sne);
}


template<class T>
inline typename List<T>::Iterator List<T>::ReverseBegin()
{
	return Iterator(sne->prev);
}

template<class T>
inline typename List<T>::ConstIterator List<T>::ReverseBegin() const
{
	return ConstIterator(sne->prev);
}


template<class T>
inline typename List<T>::Iterator List<T>::ReverseEnd()
{
	return Iterator(snb);
}

template<class T>
inline typename List<T>::ConstIterator List<T>::ReverseEnd() const
{
	return ConstIterator(snb);
}


template<class T>
inline typename List<T>::Iterator List<T>::Find(const T& t)
{
	Node *n = snb->next;
	while (n != sne)
	{
		if (n->t == t)
			return Iterator(n);

		n = n->next;
	}

	return End();
}

template<class T>
inline typename List<T>::ConstIterator List<T>::Find(const T& t) const
{
	Node *n = snb->next;
	while (n != sne)
	{
		if (n->t == t)
			return ConstIterator(n);

		n = n->next;
	}

	return End();
}


template<class T>
inline void List<T>::PushFront(const T& t)
{
	snb->t = t;
	snb->prev = new Node;
	snb->prev->next = snb;
	snb = snb->prev;

	++size;
}

template<class T>
inline void List<T>::PushBack(const T& t)
{
	sne->t = t;
	sne->next = new Node;
	sne->next->prev = sne;
	sne = sne->next;

	++size;
}


template<class T>
inline void List<T>::PopFront()
{
	assert(size > 0);

	Node *next = snb->next;
	delete snb;
	snb = next;

	--size;
}

template<class T>
inline void List<T>::PopBack()
{
	assert(size > 0);

	Node *prev = sne->prev;
	delete sne;
	sne = prev;

	--size;
}


// insert before 'it'
template<class T>
inline void List<T>::Insert(Iterator it, const T& t)
{
	Node *prev = it.n->prev;
	it.n->prev = new Node(t, prev, it.n);
	prev->next = it.n->prev;

	++size;
}


// 'it' will point to the next node
template<class T>
inline void List<T>::Erase(Iterator& it)
{
	assert(it.n != snb && it.n != sne);

	Node *next = it.n->prev->next = it.n->next;
	it.n->next->prev = it.n->prev;
	delete it.n;
	it.n = next;

	--size;
}

template<class T>
inline bool List<T>::Erase(const T& t)
{
	Node *n = snb->next;
	while (n != sne)
	{
		if (n->t == t)
		{
			n->prev->next = n->next;
			n->next->prev = n->prev;
			delete n;

			--size;

			return true;
		}

		n = n->next;
	}

	return false;
}


template<class T>
inline void List<T>::Clear()
{
	if (size > 0)
	{
		Node *c = snb->next;
		do
		{
			c = c->next;
			delete c->prev;
		}
		while (c != sne);

		snb->next = sne;
		sne->prev = snb;

		size = 0;
	}
}


template<class T>
inline List<T>::ConstIterator::ConstIterator() {
}

template<class T>
inline List<T>::ConstIterator::ConstIterator(Node *_n)
: n(_n) {
}


template<class T>
inline const T& List<T>::ConstIterator::operator* () const
{
	return n->t;
}

template<class T>
inline const T *List<T>::ConstIterator::operator-> () const
{
	return &n->t;
}


template<class T>
inline typename List<T>::ConstIterator& List<T>::ConstIterator::operator++ ()
{
	n = n->next;
	return *this;
}

template<class T>
inline typename List<T>::ConstIterator& List<T>::ConstIterator::operator-- ()
{
	n = n->prev;
	return *this;
}


template<class T>
inline bool List<T>::ConstIterator::operator== (const ConstIterator& that) const
{
	return n == that.n;
}

template<class T>
inline bool List<T>::ConstIterator::operator!= (const ConstIterator& that) const
{
	return n != that.n;
}


template<class T>
inline List<T>::Iterator::Iterator() {
}

template<class T>
inline List<T>::Iterator::Iterator(Node *_n)
: ConstIterator(_n) {
}


template<class T>
inline T& List<T>::Iterator::operator* () const
{
	return n->t;
}

template<class T>
inline T *List<T>::Iterator::operator-> () const
{
	return &n->t;
}


template<class T>
inline typename List<T>::Iterator& List<T>::Iterator::operator++ ()
{
	n = n->next;
	return *this;
}

template<class T>
inline typename List<T>::Iterator& List<T>::Iterator::operator-- ()
{
	n = n->prev;
	return *this;
}



#endif
