// Copyright (C) 2005 Christian Mendl
//
// map implementation similar to STL 'map'

#ifndef MAP_H
#define MAP_H

#include <assert.h>


template<class K, class V>
class Map
{

public:

	struct Node;

	class Iterator
	{

	public:

		Iterator();
		Iterator(Node *_n);

		V& operator* ();
		const V& operator* () const;

		V *operator-> ();
		const V *operator-> () const;

		const K& Key() const;

		Iterator& operator++ ();
		Iterator& operator-- ();

		bool operator== (const Iterator& that) const;
		bool operator!= (const Iterator& that) const;

	private:

		friend class Map;

		Node *n;
	};

	Map();
	Map(const Map& that);

	~Map();

	Map& operator= (const Map& that);

	Iterator Begin() const;
	Iterator ReverseBegin() const;
	const Iterator& End() const;

	Iterator Find(const K& k) const;

	bool Contains(const K& k) const;

	// inserts new node if key 'k' isn't in the map yet
	V& operator[] (const K& k);

	// only search for key 'k', don't insert if not in the map yet
	const V& operator[] (const K& k) const;

	int Size() const;
	bool Empty() const;

	// only inserts if k isn't in the tree yet, return value is 'ite' then;
	// otherwise pointer to existing node is returned
	Iterator Insert(const K& k, const V& v);

	// 'k' mustn't be in the tree already
	void InsertCertain(const K& k, const V& v);

	// looks for key 'k' and erases it;
	// map will remain unchanged if it doesn't contain 'k'
	void Erase(const K& k);

	void Erase(Iterator& it);

	void Clear();

private:

	static const bool red;
	static const bool black;

	struct Node
	{
		K k;
		bool c;
		Node *p, *l, *r;

		V v;

		Node(bool _c)
		: c(_c) {
		}

		Node(Node *_p)
		: p(_p) {
		}

		Node(const K& _k, bool _c, Node *_p)
		: k(_k), c(_c), p(_p), l(exn), r(exn), v() {
		}

		Node(const K& _k, bool _c, Node *_p, const V& _v)
		: k(_k), c(_c), p(_p), l(exn), r(exn), v(_v) {
		}
	};

	// sentinels

	// "sentinel root parent"
	// srp->c is black, srp->l points to the actual root and srp->p as well as srp->r to itself
	Node *srp;

	// "external node"
	// exn->c must be black, the node pointers don't matter
	static Node *exn;

	// "iterator end"
	const Iterator ite;

	int size;	// number of keys

	void RotateL(Node *n);
	void RotateR(Node *n);

	void Copy(Node *n, Node *src);

	void Erase(Node *n);
	void BalanceErase(Node *n);

	void Cleanup(Node *n);
};


template<class K, class V>
const bool Map<K, V>::red = true;

template<class K, class V>
const bool Map<K, V>::black = false;


template<class K, class V>
typename Map<K, V>::Node *Map<K, V>::exn;


template<class K, class V>
inline Map<K, V>::Map()
: size(0), ite(srp), srp(new Node(black))
{
	// we have to init 'exn' here and not where we define it; otherwise this
	// constructor might be called on a global object before 'exn' is initialized,
	// and 'exn' is reassigned after the constructor call
	if (!exn)
		exn = new Node(black);

	srp->p = srp->r = srp;
	srp->l = exn;
}

template<class K, class V>
inline Map<K, V>::Map(const Map& that)
: size(that.size), ite(srp), srp(new Node(black))
{
	srp->p = srp->r = srp;

	if (that.srp->l == exn)
		srp->l = exn;
	else
	{
		srp->l = new Node(srp);
		Copy(srp->l, that.srp->l);
	}
}


template<class K, class V>
inline Map<K, V>::~Map()
{
	if (srp->l != exn)
		Cleanup(srp->l);

	delete srp;
}


template<class K, class V>
inline Map<K, V>& Map<K, V>::operator= (const Map& that)
{
	if (srp->l != exn)
		Cleanup(srp->l);

	size = that.size;

	if (that.srp->l == exn)
		srp->l = exn;
	else
	{
		srp->l = new Node(srp);
		Copy(srp->l, that.srp->l);
	}

	return *this;
}


template<class K, class V>
inline typename Map<K, V>::Iterator Map<K, V>::Begin() const
{
	Node *n = srp;

	while (n->l != exn)
		n = n->l;

	return Iterator(n);
}

template<class K, class V>
inline typename Map<K, V>::Iterator Map<K, V>::ReverseBegin() const
{
	if (srp->l == exn)
		return ite;

	Node *n = srp->l;

	while (n->r != exn)
		n = n->r;

	return Iterator(n);
}

template<class K, class V>
inline const typename Map<K, V>::Iterator& Map<K, V>::End() const
{
	return ite;
}


template<class K, class V>
inline typename Map<K, V>::Iterator Map<K, V>::Find(const K& k) const
{
	Node *n = srp->l;

	while (n != exn)
	{
		if (k < n->k)
			n = n->l;
		else if (n->k < k)
			n = n->r;
		else
			return Iterator(n);
	}

	return ite;
}


template<class K, class V>
inline bool Map<K, V>::Contains(const K& k) const
{
	Node *n = srp->l;

	while (n != exn)
	{
		if (k < n->k)
			n = n->l;
		else if (n->k < k)
			n = n->r;
		else
			return true;
	}

	return false;
}


template<class K, class V>
inline V& Map<K, V>::operator[] (const K& k)
{
	if (srp->l != exn)
	{
		Node *n, *w = srp->l;

		while (true)
		{
			if (k < w->k)
			{
				if (w->l != exn)
					w = w->l;
				else
				{
					w->l = n = new Node(k, red, w);
					w = n;
					break;
				}
			}
			else if (w->k < k)
			{
				if (w->r != exn)
					w = w->r;
				else
				{
					w->r = n = new Node(k, red, w);
					w = n;
					break;
				}
			}
			else
				return w->v;
		}

		while (n->p->c == red)
		{
			if (n->p == n->p->p->l)
			{
				Node *m = n->p->p->r;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->r)
						RotateL(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateR(n->p);

					break;
				}
			}
			else
			{
				Node *m = n->p->p->l;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->l)
						RotateR(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateL(n->p);

					break;
				}
			}
		}

		srp->l->c = black;

		++size;

		return w->v;
	}
	else
	{
		srp->l = new Node(k, black, srp);

		++size;

		return srp->l->v;
	}
}


template<class K, class V>
inline const V& Map<K, V>::operator[] (const K& k) const
{
	Node *n = srp->l;

	while (n != exn)
	{
		if (k < n->k)
			n = n->l;
		else if (n->k < k)
			n = n->r;
		else
			break;
	}

	return n->v;
}


template<class K, class V>
inline int Map<K, V>::Size() const
{
	return size;
}

template<class K, class V>
inline bool Map<K, V>::Empty() const
{
	return size == 0;
}


template<class K, class V>
inline typename Map<K, V>::Iterator Map<K, V>::Insert(const K& k, const V& v)
{
	if (srp->l != exn)
	{
		Node *n = srp->l;

		while (true)
		{
			if (k < n->k)
			{
				if (n->l != exn)
					n = n->l;
				else
				{
					n->l = new Node(k, red, n, v);
					n = n->l;
					break;
				}
			}
			else if (n->k < k)
			{
				if (n->r != exn)
					n = n->r;
				else
				{
					n->r = new Node(k, red, n, v);
					n = n->r;
					break;
				}
			}
			else
				return Iterator(n);
		}

		while (n->p->c == red)
		{
			if (n->p == n->p->p->l)
			{
				Node *m = n->p->p->r;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->r)
						RotateL(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateR(n->p);

					break;
				}
			}
			else
			{
				Node *m = n->p->p->l;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->l)
						RotateR(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateL(n->p);

					break;
				}
			}
		}

		srp->l->c = black;
	}
	else
		srp->l = new Node(k, black, srp, v);

	size++;

	return ite;
}

template<class K, class V>
inline void Map<K, V>::InsertCertain(const K& k, const V& v)
{
	assert(Find(k) == ite);

	if (srp->l != exn)
	{
		Node *n = srp->l;

		while (true)
		{
			if (k < n->k)
			{
				if (n->l != exn)
					n = n->l;
				else
				{
					n->l = new Node(k, red, n, v);
					n = n->l;
					break;
				}
			}
			else
			{
				if (n->r != exn)
					n = n->r;
				else
				{
					n->r = new Node(k, red, n, v);
					n = n->r;
					break;
				}
			}
		}

		while (n->p->c == red)
		{
			if (n->p == n->p->p->l)
			{
				Node *m = n->p->p->r;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->r)
						RotateL(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateR(n->p);

					break;
				}
			}
			else
			{
				Node *m = n->p->p->l;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->l)
						RotateR(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateL(n->p);

					break;
				}
			}
		}

		srp->l->c = black;
	}
	else
		srp->l = new Node(k, black, srp, v);

	size++;
}


template<class K, class V>
inline void Map<K, V>::Erase(const K& k)
{
	// 'size' may be 0

	Node *n = srp->l;

	while (n != exn)
	{
		if (k < n->k)
			n = n->l;
		else if (n->k < k)
			n = n->r;
		else
		{
			Erase(n);
			return;
		}
	}
}

template<class K, class V>
inline void Map<K, V>::Erase(Iterator& it)
{
	assert(size > 0);

	Erase(it.n);
}


template<class K, class V>
inline void Map<K, V>::Clear()
{
	if (srp->l != exn)
	{
		Cleanup(srp->l);
		srp->l = exn;
		size = 0;
	}
}


template<class K, class V>
inline void Map<K, V>::RotateL(Node *n)
{
	Node *m = n->r;

	n->r = m->l;
	n->r->p = n;	// unconditional assignment doesn't introduce a bug as long as we don't care where exn->p points to

	m->p = n->p;
	if (n == n->p->l)
		n->p->l = m;
	else
		n->p->r = m;

	m->l = n;
	n->p = m;
}

template<class K, class V>
inline void Map<K, V>::RotateR(Node *n)
{
	Node *m = n->l;

	n->l = m->r;
	n->l->p = n;

	m->p = n->p;
	if (n == n->p->l)
		n->p->l = m;
	else
		n->p->r = m;

	m->r = n;
	n->p = m;
}


// can't be inline
template<class K, class V>
void Map<K, V>::Copy(Node *n, Node *src)
{
	n->k = src->k;
	n->c = src->c;
	n->v = src->v;

	if (src->l == exn)
		n->l = exn;
	else
	{
		n->l = new Node(n);
		Copy(n->l, src->l);
	}

	if (src->r == exn)
		n->r = exn;
	else
	{
		n->r = new Node(n);
		Copy(n->r, src->r);
	}
}


template<class K, class V>
inline void Map<K, V>::Erase(Node *n)
{
	assert(size > 0);

	if (n->l == exn)
	{
		n->r->p = n->p;

		if (n == n->p->l)
			n->p->l = n->r;
		else
			n->p->r = n->r;

		if (n->c == black)
			BalanceErase(n->r);
	}
	else if (n->r == exn)
	{
		n->l->p = n->p;

		if (n == n->p->l)
			n->p->l = n->l;
		else
			n->p->r = n->l;

		if (n->c == black)
			BalanceErase(n->l);
	}
	else
	{
		Node *m = n->r;

		if (m->l == exn)
		{
			n->l->p = m;

			if (n == n->p->l)
				n->p->l = m;
			else
				n->p->r = m;

			m->p = n->p;
			m->l = n->l;

			if (m->c == black)
			{
				m->c = n->c;
				m->r->p = m;	// make sure that exn has parent m if m->r == exn
				BalanceErase(m->r);
			}
			else
				m->c = n->c;
		}
		else
		{
			m = m->l;
			while (m->l != exn)
				m = m->l;

			n->l->p = m;
			n->r->p = m;

			if (n == n->p->l)
				n->p->l = m;
			else
				n->p->r = m;

			m->r->p = m->p;
			m->p->l = m->r;	// m is a left child

			m->p = n->p;
			m->l = n->l;

			if (m->c == black)
			{
				m->c = n->c;
				Node *v = m->r;
				m->r = n->r;
				BalanceErase(v);
			}
			else
			{
				m->c = n->c;
				m->r = n->r;
			}
		}
	}

	delete n;

	size--;
}

template<class K, class V>
inline void Map<K, V>::BalanceErase(Node *n)
{
	if (n->c == red)
	{
		n->c = black;
		return;
	}

	// use the srp sentinel instead of checking for root
	while (true)
	{
		if (n == n->p->l)
		{
			// if we didn't assign n to its parent and n == exn, we'll run into an assertion as exn->p is changed by RotateL/R functions
			n = n->p;

			Node *m = n->r;

			if (m->c == red)
			{
				m->c = black;
				n->c = red;

				RotateL(n);

				m = n->r;
			}

			if (m->r->c == red)
			{
				m->r->c = black;
				m->c = n->c;
				n->c = black;

				RotateL(n);

				return;
			}
			else if (m->l->c == red)
			{
				m->l->c = black;
				m->c = red;

				RotateR(m);

				m->p->c = n->c;
				n->c = black;
				m->c = black;

				RotateL(n);

				return;
			}
			else
			{
				m->c = red;

				if (n->c == red)
				{
					n->c = black;
					return;
				}
			}
		}
		else
		{
			n = n->p;

			Node *m = n->l;

			if (m->c == red)
			{
				m->c = black;
				n->c = red;

				RotateR(n);

				m = n->l;
			}

			if (m->l->c == red)
			{
				m->l->c = black;
				m->c = n->c;
				n->c = black;

				RotateR(n);

				return;
			}
			else if (m->r->c == red)
			{
				m->r->c = black;
				m->c = red;

				RotateL(m);

				m->p->c = n->c;
				n->c = black;
				m->c = black;

				RotateR(n);

				return;
			}
			else
			{
				m->c = red;

				if (n->c == red)
				{
					n->c = black;
					return;
				}
			}
		}
	}
}


// can't be inline
template<class K, class V>
void Map<K, V>::Cleanup(Node *n)
{
	if (n->l != exn)
		Cleanup(n->l);

	if (n->r != exn)
		Cleanup(n->r);

	delete n;
}


template<class K, class V>
inline Map<K, V>::Iterator::Iterator() {
}

template<class K, class V>
inline Map<K, V>::Iterator::Iterator(Node *_n)
: n(_n) {
}


template<class K, class V>
inline V& Map<K, V>::Iterator::operator* ()
{
	return n->v;
}

template<class K, class V>
inline const V& Map<K, V>::Iterator::operator* () const
{
	return n->v;
}


template<class K, class V>
inline V *Map<K, V>::Iterator::operator-> ()
{
	return &n->v;
}

template<class K, class V>
inline const V *Map<K, V>::Iterator::operator-> () const
{
	return &n->v;
}


template<class K, class V>
inline const K& Map<K, V>::Iterator::Key() const
{
	return n->k;
}


template<class K, class V>
inline typename Map<K, V>::Iterator& Map<K, V>::Iterator::operator++ ()
{
	if (n->r != exn)
	{
		n = n->r;

		while (n->l != exn)
			n = n->l;
	}
	else
	{
		while (n == n->p->r)
			n = n->p;

		n = n->p;
	}

	return *this;
}

template<class K, class V>
inline typename Map<K, V>::Iterator& Map<K, V>::Iterator::operator-- ()
{
	if (n->l != exn)
	{
		n = n->l;

		while (n->r != exn)
			n = n->r;
	}
	else
	{
		while (n == n->p->l)
			n = n->p;

		n = n->p;
	}

	return *this;
}


template<class K, class V>
inline bool Map<K, V>::Iterator::operator== (const Iterator& that) const
{
	return n == that.n;
}

template<class K, class V>
inline bool Map<K, V>::Iterator::operator!= (const Iterator& that) const
{
	return n != that.n;
}



#endif
