// Copyright (C) 2005 Christian Mendl
//
// behaves like an array, but elements are only inserted if not in the array yet;
// implements a red-black-tree

#ifndef UNIQUE_H
#define UNIQUE_H

#include <new>
#include <assert.h>
#include "next2k.h"


template<class T>
class Unique
{

public:

	Unique();
	Unique(const Unique& that);

	~Unique();

	Unique& operator= (const Unique& that);

	const T& operator[] (int i) const;

	const T* Data() const;

	int Size() const;
	bool Empty() const;

	// number of bytes stored in the data array
	int NumBytes() const;

	const T& Front() const;
	const T& Back() const;

	// if 't' is in array, return index; -1 otherwise
	int Find(const T& t) const;

	// element is only inserted if not in the array yet;
	// return element index
	int PushBack(const T& t);

	template<template<class> class Ar>
	void PushBack(const Ar<T>& that);

	void PopBack();

	// erase element at index i;
	// alternatively use a linked list when inserting and erasing elements often
	void EraseAt(int i);

	// erase element 't'; return false if 't' isn't in the array
	bool Erase(const T& t);

	void Shrink(int ns);

	void Clear();

private:

	struct Node
	{
		int i;				// array element index
		bool c;				// color
		Node *p, *l, *r;	// parent, left child, right child

		Node(bool _c)
		: c(_c) {
		}

		Node(Node *_p)
		: p(_p) {
		}

		Node(int _i, bool _c, Node *_p)
		: i(_i), c(_c), p(_p), l(exn), r(exn) {
		}
	};

	static const bool red;
	static const bool black;

	T *data;		// data array pointer
	Node **nodes;	// node pointer

	int size, capacity;	// capacity always 2^k for a number k >= 2

	// sentinels

	// "sentinel root parent"
	// srp->c is black, srp->l points to the actual root and srp->p as well as srp->r to itself
	Node *srp;

	// "external node"
	// exn->c must be black, the node pointers don't matter
	// must be static to be used in 'Node' constructor and in 'Copy'
	static Node *exn;

	void RotateL(Node *n);
	void RotateR(Node *n);

	// only inserts if t isn't in the tree yet;
	Node *InsertNode(const T& t);

	void Copy(Node *n, Node *src);

	void EraseNode(Node *n);
	void BalanceErase(Node *n);

	void Cleanup(Node *n);
};


template<class T>
const bool Unique<T>::red = true;

template<class T>
const bool Unique<T>::black = false;


template<class T>
typename Unique<T>::Node *Unique<T>::exn;


template<class T>
inline Unique<T>::Unique()
: size(0), srp(new Node(black))
{
	// we have to init 'exn' here and not where we define it because otherwise,
	// this constructor might be called on an global object before 'exn' is initialized,
	// and 'exn' is reassigned after the constructor call
	if (!exn)
		exn = new Node(black);

	capacity = 4;
	data = (T *)malloc(capacity*sizeof(T));
	nodes = (Node **)malloc(capacity*sizeof(Node *));

	srp->p = srp->r = srp;
	srp->l = exn;
}

template<class T>
inline Unique<T>::Unique(const Unique& that)
{
	size = that.size;

	capacity = that.capacity;
	data = (T *)malloc(capacity*sizeof(T));
	nodes = (Node **)malloc(capacity*sizeof(Node *));

	for (int i = 0; i < size; i++)
		new(data+i) T(that.data[i]);

	srp = new Node(black);
	srp->p = srp->r = srp;

	if (that.srp->l == exn)
		srp->l = exn;
	else
	{
		srp->l = new Node(srp);
		Copy(srp->l, that.srp->l);
	}
}


template<class T>
inline Unique<T>::~Unique()
{
	if (srp->l != exn)
		Cleanup(srp->l);
	delete srp;

	free(nodes);

	for (int i = size-1; i >= 0; i--)
		(data+i)->~T();
	free(data);
}


template<class T>
inline Unique<T>& Unique<T>::operator= (const Unique& that)
{
	if (srp->l != exn)
		Cleanup(srp->l);

	for (int i = size-1; i >= 0; i--)
		(data+i)->~T();

	size = that.size;

	if (capacity < that.capacity)
	{
		capacity = that.capacity;

		// this should be faster than realloc

		free(nodes);
		free(data);

		data = (T *)malloc(capacity*sizeof(T));
		nodes = (Node **)malloc(capacity*sizeof(Node *));
	}

	for (int i = 0; i < that.size; i++)
		new(data+i) T(that.data[i]);

	if (that.srp->l == exn)
		srp->l = exn;
	else
	{
		srp->l = new Node(srp);
		Copy(srp->l, that.srp->l);
	}

	return *this;
}


template<class T>
inline const T& Unique<T>::operator[] (int i) const
{
	assert(0 <= i && i < size);

	return data[i];
}


template<class T>
inline const T* Unique<T>::Data() const
{
	return data;
}


template<class T>
inline int Unique<T>::Size() const
{
	return size;
}

template<class T>
inline bool Unique<T>::Empty() const
{
	return size == 0;
}

template<class T>
inline int Unique<T>::NumBytes() const
{
	return size * sizeof(T);
}


template<class T>
inline const T& Unique<T>::Front() const
{
	assert(size > 0);

	return data[0];
}

template<class T>
inline const T& Unique<T>::Back() const
{
	assert(size > 0);

	return data[size-1];
}


template<class T>
inline int Unique<T>::Find(const T& t) const
{
	Node *n = srp->l;

	while (n != exn)
	{
		if (t < data[n->i])
			n = n->l;
		else if (data[n->i] < t)
			n = n->r;
		else
			return n->i;
	}

	return -1;
}


template<class T>
inline int Unique<T>::PushBack(const T& t)
{
	Node *n = InsertNode(t);

	if (n->i == size)
	{
		if (capacity < size + 1)
		{
			capacity *= 2;

			data = (T *)realloc(data, capacity*sizeof(T));
			nodes = (Node **)realloc(nodes, capacity*sizeof(Node *));
		}

		new(data + size) T(t);
		nodes[size] = n;

		return size++;
	}
	else
		return n->i;
}


template<class T>
template<template<class> class Ar>
inline void Unique<T>::PushBack(const Ar<T>& that)
{
	assert(this != &that);

	if (capacity < size + that.Size())
	{
		capacity = Next2K(size + that.Size());

		data = (T *)realloc(data, capacity*sizeof(T));
		nodes = (Node **)realloc(nodes, capacity*sizeof(Node *));
	}

	for (int i = 0; i < that.Size(); i++)
	{
		Node *n = InsertNode(that[i]);

		if (n->i == size)
		{
			new(data + size) T(that[i]);
			nodes[size] = n;

			++size;
		}
	}
}


template<class T>
inline void Unique<T>::PopBack()
{
	assert(size > 0);

	--size;

	EraseNode(nodes[size]);
	(data+size)->~T();
}


template<class T>
inline void Unique<T>::EraseAt(int i)
{
	assert(0 <= i && i < size);

	EraseNode(nodes[i]);

	--size;

	// shift elements
	for (int k = i; k < size; k++)
	{
		data[k] = data[k+1];
		nodes[k] = nodes[k+1];

		--nodes[k]->i;
	}

	// delete last element
	(data+size)->~T();
}

template<class T>
inline bool Unique<T>::Erase(const T& t)
{
	Node *n = srp->l;

	while (n != exn)
	{
		if (t < data[n->i])
			n = n->l;
		else if (data[n->i] < t)
			n = n->r;
		else
		{
			--size;

			// shift elements
			for (int i = n->i; i < size; i++)
			{
				data[i] = data[i+1];
				nodes[i] = nodes[i+1];

				--nodes[i]->i;
			}

			// delete last element
			(data+size)->~T();

			EraseNode(n);

			return true;
		}
	}

	return false;
}


template<class T>
inline void Unique<T>::Shrink(int ns)
{
	assert(ns >= 0 && size-ns >= 0);

	for (int i = size-1; i >= size-ns; i--)
	{
		EraseNode(nodes[i]);
		(data+i)->~T();
	}

	size -= ns;
}


template<class T>
inline void Unique<T>::Clear()
{
	if (srp->l != exn)
	{
		Cleanup(srp->l);
		srp->l = exn;

		for (int i = size-1; i >= 0; i--)
			(data+i)->~T();

		size = 0;
	}
}


template<class T>
inline void Unique<T>::RotateL(Node *n)
{
	Node *m = n->r;

	n->r = m->l;
	n->r->p = n;	// unconditional assignment doesn't introduce a bug as long as we don't care where exn->p points to

	m->p = n->p;
	if (n == n->p->l)
		n->p->l = m;
	else
		n->p->r = m;

	m->l = n;
	n->p = m;
}

template<class T>
inline void Unique<T>::RotateR(Node *n)
{
	Node *m = n->l;

	n->l = m->r;
	n->l->p = n;

	m->p = n->p;
	if (n == n->p->l)
		n->p->l = m;
	else
		n->p->r = m;

	m->r = n;
	n->p = m;
}


template<class T>
inline typename Unique<T>::Node *Unique<T>::InsertNode(const T& t)
{
	if (srp->l != exn)
	{
		Node *n, *w = srp->l;

		while (true)
		{
			if (t < data[w->i])
			{
				if (w->l != exn)
					w = w->l;
				else
				{
					w->l = n = new Node(size, red, w);
					w = n;
					break;
				}
			}
			else if (data[w->i] < t)
			{
				if (w->r != exn)
					w = w->r;
				else
				{
					w->r = n = new Node(size, red, w);
					w = n;
					break;
				}
			}
			else
				return w;
		}

		while (n->p->c == red)
		{
			if (n->p == n->p->p->l)
			{
				Node *m = n->p->p->r;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->r)
						RotateL(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateR(n->p);

					break;
				}
			}
			else
			{
				Node *m = n->p->p->l;

				if (m->c == red)
				{
					n->p->c = m->c = black;
					n = m->p;
					n->c = red;
				}
				else
				{
					if (n == n->p->l)
						RotateR(n->p);
					else
						n = n->p;

					n->c = black;
					n->p->c = red;

					RotateL(n->p);

					break;
				}
			}
		}

		srp->l->c = black;

		return w;
	}
	else
	{
		srp->l = new Node(size, black, srp);

		return srp->l;
	}
}


// can't be inline
template<class T>
void Unique<T>::Copy(Node *n, Node *src)
{
	n->i = src->i;
	n->c = src->c;

	nodes[n->i] = n;

	if (src->l == exn)
		n->l = exn;
	else
	{
		n->l = new Node(n);
		Copy(n->l, src->l);
	}

	if (src->r == exn)
		n->r = exn;
	else
	{
		n->r = new Node(n);
		Copy(n->r, src->r);
	}
}


template<class T>
inline void Unique<T>::EraseNode(Node *n)
{
	if (n->l == exn)
	{
		n->r->p = n->p;

		if (n == n->p->l)
			n->p->l = n->r;
		else
			n->p->r = n->r;

		if (n->c == black)
			BalanceErase(n->r);
	}
	else if (n->r == exn)
	{
		n->l->p = n->p;

		if (n == n->p->l)
			n->p->l = n->l;
		else
			n->p->r = n->l;

		if (n->c == black)
			BalanceErase(n->l);
	}
	else
	{
		Node *m = n->r;

		if (m->l == exn)
		{
			n->l->p = m;

			if (n == n->p->l)
				n->p->l = m;
			else
				n->p->r = m;

			m->p = n->p;
			m->l = n->l;

			if (m->c == black)
			{
				m->c = n->c;
				m->r->p = m;	// make sure that exn has parent m if m->r == exn
				BalanceErase(m->r);
			}
			else
				m->c = n->c;
		}
		else
		{
			m = m->l;
			while (m->l != exn)
				m = m->l;

			n->l->p = m;
			n->r->p = m;

			if (n == n->p->l)
				n->p->l = m;
			else
				n->p->r = m;

			m->r->p = m->p;
			m->p->l = m->r;	// m is a left child

			m->p = n->p;
			m->l = n->l;

			if (m->c == black)
			{
				m->c = n->c;
				Node *v = m->r;
				m->r = n->r;
				BalanceErase(v);
			}
			else
			{
				m->c = n->c;
				m->r = n->r;
			}
		}
	}

	delete n;
}

template<class T>
inline void Unique<T>::BalanceErase(Node *n)
{
	if (n->c == red)
	{
		n->c = black;
		return;
	}

	// use the srp sentinel instead of checking for root
	while (true)
	{
		if (n == n->p->l)
		{
			// if we didn't assign n to its parent and n == exn, we'll run into an assertion as exn->p is changed by RotateL/R functions
			n = n->p;

			Node *m = n->r;

			if (m->c == red)
			{
				m->c = black;
				n->c = red;

				RotateL(n);

				m = n->r;
			}

			if (m->r->c == red)
			{
				m->r->c = black;
				m->c = n->c;
				n->c = black;

				RotateL(n);

				return;
			}
			else if (m->l->c == red)
			{
				m->l->c = black;
				m->c = red;

				RotateR(m);

				m->p->c = n->c;
				n->c = black;
				m->c = black;

				RotateL(n);

				return;
			}
			else
			{
				m->c = red;

				if (n->c == red)
				{
					n->c = black;
					return;
				}
			}
		}
		else
		{
			n = n->p;

			Node *m = n->l;

			if (m->c == red)
			{
				m->c = black;
				n->c = red;

				RotateR(n);

				m = n->l;
			}

			if (m->l->c == red)
			{
				m->l->c = black;
				m->c = n->c;
				n->c = black;

				RotateR(n);

				return;
			}
			else if (m->r->c == red)
			{
				m->r->c = black;
				m->c = red;

				RotateL(m);

				m->p->c = n->c;
				n->c = black;
				m->c = black;

				RotateR(n);

				return;
			}
			else
			{
				m->c = red;

				if (n->c == red)
				{
					n->c = black;
					return;
				}
			}
		}
	}
}


// can't be inline
template<class T>
void Unique<T>::Cleanup(Node *n)
{
	if (n->l != exn)
		Cleanup(n->l);

	if (n->r != exn)
		Cleanup(n->r);

	delete n;
}



#endif
