// Copyright (C) 2005 Christian Mendl

#include "hash_set.h"
#include <iostream>
using namespace std;


template<class K, class H>
void Print(const HashSet<K, H>& hs)
{
	cout << "Size: " << hs.Size() << endl;
	if (hs.Empty())
		cout << "empty." << endl;
	else
	{
		HashSet<K, H>::Iterator it;

		cout << "Elements: ";
		for (it = hs.Begin(); it != hs.End(); ++it)
			cout << *it << " ";

		cout << endl << "Reverse: ";
		for (it = hs.ReverseBegin(); it != hs.ReverseEnd(); --it)
			cout << *it << " ";

		cout << endl;
	}
}

template<class K, class H>
void Search(const HashSet<K, H>& hs, const K& k)
{
	cout << endl << "Looking for " << k << "..." << endl;
	HashSet<K, H>::Iterator it = hs.Find(k);
	if (it == hs.End())
	{
		assert(!hs.Contains(k));
		cout << k << " not in hash set." << endl;
	}
	else
	{
		assert(hs.Contains(k));
		cout << k << " is in hash set." << endl;
	}
}


class Key
{

public:

	Key()
	{
		data = new double[2];
		data[0] = data[1] = 0;

		++numAlloc;
	}

	Key(double a, double b)
	{
		data = new double[2];
		data[0] = a;
		data[1] = b;

		++numAlloc;
	}

	Key(const Key& that)
	{
		data = new double[2];
		data[0] = that.data[0];
		data[1] = that.data[1];

		++numAlloc;
	}

	Key& operator= (const Key& that)
	{
		data[0] = that.data[0];
		data[1] = that.data[1];

		return *this;
	}

	~Key()
	{
		--numAlloc;

		delete [] data;
	}

	double operator[] (int i) const
	{
		assert(0 <= i && i < 2);
		return data[i];
	}

	bool operator== (const Key& that) const
	{
		return data[0] == that.data[0] && data[1] == that.data[1];
	}

	bool operator!= (const Key& that) const
	{
		return data[0] != that.data[0] || data[1] != that.data[1];
	}

	bool operator< (const Key& that) const
	{
		if (data[0] < that.data[0])
			return true;
		else if (that.data[0] < data[0])
			return false;

		return data[1] < that.data[1];
	}

	static int numAlloc;

private:

	double *data;
};

int Key::numAlloc = 0;

ostream& operator<< (ostream& strm, const Key& key)
{
	strm << "(" << key[0] << ", " << key[1] << ")";
	return strm;
}


struct Hash
{
	int operator() (const Key& key) const
	{
		return abs((int)key[0]);
	}
};


void Test()
{
	cout << endl << "Creating hash set and inserting some elements..." << endl;
	HashSet<Key, Hash> hs;
	hs.Insert(Key(1.23, 0));
	hs.Insert(Key(-1, 1.24));
	hs.Insert(Key(-1, 2.5));
	Print(hs);

	cout << endl << "Copy constructor..." << endl;
	HashSet<Key, Hash> hs2(hs);
	Print(hs2);

	cout << endl << "Inserting some more elements..." << endl;
	hs.Insert(Key(77.7, 666));
	hs.Insert(Key(123, -123));
	Print(hs);

	Search(hs, Key(6, 0));
	Search(hs, Key(-1, 1.24));

	cout << endl << "Trying to erase (-1, 2.4) and (-1, 2.5)..." << endl;
	hs.Erase(Key(-1, 2.4));
	hs.Erase(Key(-1, 2.5));
	Print(hs);

	cout << endl << "Erasing last element..." << endl;
	hs.Erase(hs.ReverseBegin());
	Print(hs);

	cout << endl << "Clearing second hash set..." << endl;
	hs2.Clear();
	Print(hs2);

	cout << endl << "Assigning first to second..." << endl;
	hs2 = hs;
	Print(hs2);

	cout << endl << "Passing capacity to copy constructor..." << endl;
	HashSet<Key, Hash> hs3(16);
	hs3.Insert(Key(-1, 1));
	hs3.Insert(Key(0, 0));
	hs3.Insert(Key(24, 10));
	Print(hs3);
}

void main()
{
	cout << "Key count: " << Key::numAlloc << endl;

	// objects are destructed on function call return
	Test();

	cout << endl << "Key count: " << Key::numAlloc << endl;
}
